//  $Id: prom_elem.hh,v 1.2 2004/06/16 00:12:24 adams Exp $
//  Author:             Mark F. Adams
// Copyright (c) 2000 by Mark F. Adams 
//  Filename:           prom_elem.hh
//
#ifndef __ELEMENT_H__
#define __ELEMENT_H__

#include "prom_node.hh"

//! class PromMaterial: FE material 
/*!
  class PromMaterial: FE material 
*/
class PromMaterial
{
public:
  PromMaterial( int id = -9 ){ matID_ = id; }
  //virtual ~PromMaterial(){};
  int  getMaterialID()const{ assert(this!=NULL); return matID_; }
  void setMaterialID(int id){ assert(this!=NULL); matID_=id;}
  void setMaterialID(PromMaterial &a){ operator=(a); }
  PromMaterial &operator=( const PromMaterial &a ) { 
    assert(this!=NULL); 
    matID_ = a.matID_;
    return *this; 
  }
  int Write( int **ppb )const{ 
    int *pb = *ppb; *pb++ = matID_; *ppb = pb; return 0; 
  }
  int Read( int **ppb ){ int *pb = *ppb; matID_ = *pb++; *ppb = pb; return 0; }
  static int GetWrite_size( int &sz ) { sz = 1; return 0; }
private:
  int  matID_;
};

//! class PromPoly: Bsse class for polyhedron 
/*!
  class PromPoly: Bsse class for polyhedron -- 3D
*/
class PromPoly : public PromElement, public  PromMaterial, public PromStack    
{
public:  
  PromPoly( PromNode **n = NULL, PromPoly **a = NULL, int matid = 0)
    : PromMaterial(matid) { nodes_ = n; neigh_ = a;}
  //virtual ~PromPoly(){};
  void reset( const int numnp, PromNode *a[] );
  virtual void reset(){ 
    bits_.done = bits_.qual = 0;  bits_.dead = 1; 
    assert(isChildListEmpty());
  }
  virtual int numNodes() const = 0;
  virtual int numNeigh() const = 0; 
  virtual const int * getIDArr()const = 0;
  virtual int numGhosts() const{ 
    assert(this!=NULL);  int ii, cc;
    for ( ii = 0, cc = 0 ; ii < numNodes() ; ii++ ) 
      if( nodes_[ii]->isGhost() ) cc++;
    return cc;
  }
  virtual int numExposedFaces()const{
    assert(this!=NULL);  int ii, cc;
    for ( ii = 0, cc = 0 ; ii < numNeigh() ; ii++ ) if ( neigh_[ii]==0 ) cc++;
    return cc;
  }
  virtual int WriteFEAP( FILE *file, const int pg[] = NULL ) const;
  virtual int numEnv() const{ 
    assert(this!=NULL); int ii, cc;
    for ( ii = 0, cc = 0 ; ii < numNodes() ; ii++ ) 
      if( nodes_[ii]->isEnvelope() ) cc++;
    return cc;
  }  
  int GetNodes( PromTable<PromNode*> *hull )const {
    assert(this!=NULL);
    for( int j = 0 ; j < numNodes() ; j++ ) hull->add( nodes_[j] );   
    return 0;
  }
  int AddSelfToNodes() { 
    assert(this!=NULL);
    for(int j=0 ; j < numNodes() ; j++) {
      PromNode *node = nodes_[j];
      node->symbData_->elemList.addTail( this );
    }
    return 0;
  }
  int AddAllToAllAdjac()const { 
    for(int i = 0 ; i < numNodes() ; i++ ) {
      for(int j = 0 ; j < numNodes() ; j++ ) {
	if(i!=j){
	  assert(nodes_[i] != nodes_[j]); 
	  nodes_[i]->symbData_->adjac.add(nodes_[j]);
	}
      }
    }
    return 0; 
  }
  // data
  PromNode  **nodes_;    
  PromPoly  **neigh_;    
  // global helper
  static const int tetID_[4][4];
  static const int octID_[6][4];
  static const int pyramidID_[5][4];
  static const int wedgeID_[5][4];
  static const int septID_[6][4];
  static const int quadID_[4][4];
};

//! class PromDTet: Delaunay tetrahedron
/*!
  class PromDTet: Delaunay tetrahedron
*/
class PromDTet : public PromPoly
{
public:
  PromDTet() : PromPoly() {
    setType(PROM_SOLID);
    nodes_ = nodes_private_;
    neigh_ = (PromPoly**)neigh_private_; bits_.dead = 1;
    reset();
    assert( isDead() );
  }
  PromDTet( PromNode *a, PromNode *b, PromNode *c, PromNode *d){ 
    nodes_ = nodes_private_;
    neigh_ = (PromPoly**)neigh_private_;
    reset( a, b, c, d );
  }
  void reset( PromNode *a, PromNode *b, PromNode *c, PromNode *d );
  //virtual ~PromDTet(){}
  virtual void reset();
  PromDTet& operator=( const PromDTet& a );

  virtual int numNodes() const{ assert(this!=NULL); return 4;}
  virtual int numNeigh() const{ assert(this!=NULL); return 4;}// even if void ?
  const int * getIDArr()const{ return tetID_[0]; }
  
  int numEnv() const{
    assert(this!=NULL);
    return nodes_[0]->isEnvelope() + nodes_[1]->isEnvelope() +
      nodes_[2]->isEnvelope() + nodes_[3]->isEnvelope() ;
  }
  int numExterior() const{ 
    assert(this!=NULL);
    return nodes_[0]->isExterior() + nodes_[1]->isExterior() +
      nodes_[2]->isExterior() + nodes_[3]->isExterior() ;
  }
  int numExposedFaces()const{
    assert(this!=NULL);
    return (!neigh_[0] + !neigh_[1] + !neigh_[2] + !neigh_[3]);
  }
  int numGhosts()const{
    assert(this!=NULL);
    return nodes_[0]->isGhost() + nodes_[1]->isGhost() + 
      nodes_[2]->isGhost() + nodes_[3]->isGhost();
  }
  bool willStrandNode( int &ret )const;

  // generation functions TET
  gfloat calcTestData();
  bool   isInCirc( PromCoord &c, double* );
  bool   isIn(const PromCoord &c,shpfloat s[],const gfloat t,int mype,bool*ok);
  int    SetNeigh( PromDTet *newel, const PromNode *n1, const PromNode *n2,
		   const PromNode *n3, const PromNode *n4 = NULL );
  shpfloat calcQuality();
  gfloat   getVolume() { assert(this!=NULL); return tdttVol_; }
  bool     isSliver();
  shpfloat quality(){ 
    assert( this != NULL );
    if ( bits_.qual==0 ) { calcQuality(); bits_.qual = 1; } 
    return quality_; 
  }
  static bool shape( PromNode *n[], const PromCoord &crd, shpfloat vs[]);
private:
  PromNode   *nodes_private_[4];
  PromDTet   *neigh_private_[4];
  gfloat      tdttVol_; 
  gfloat      quality_;
};

//! class PromFacet: facet (of a polyhedron 
/*!
  class PromFacet: facet (of a polyhedron -- 2D
*/
class PromFacet : public PromPoly
{
public:
  // constructor
  PromFacet() : PromPoly() {
    nodes_ = (PromNode**)&nodes_private_;
    neigh_ = (PromPoly**)&neigh_private_;
    reset(); 
    assert( isDead() );
  }
  PromFacet( PromNode *n1, PromNode *n2, PromNode *n3, 
	     PromNode *n4 = NULL) {
    nodes_ = (PromNode**)nodes_private_;
    neigh_ = (PromPoly**)neigh_private_;
    reset( n1, n2, n3, n4);
  }
  void reset( PromNode *n1, PromNode *n2, PromNode *n3, 
	      PromNode *n4 = NULL );
  virtual void  reset();
  const int * getIDArr()const{ assert(0); return NULL; }
  virtual int numNodes()const{ assert(this!=NULL); return nodes_[3] ? 4 : 3;}
  virtual int numNeigh()const{ assert(this!=NULL); return neigh_[3] ? 4 : 3;}
  
  int         getID() const{ return face_; }
  void        setFaceID(int id){ assert(id>0); face_ = id; }
    
  /*bool rightSide( const PromGeoNode *node, gfloat *ret = NULL )const {
    assert(this!=NULL);
    PromCoord &crd0 = nodes_private_[0]->symbGeoData_->coord;
    PromCoord &crd = node->symbGeoData_->coord;
    PromCoord vec(crd[0] - crd0[0], crd[1] - crd0[1], crd[2] - crd0[2]);  
    vec /= vec.length();
    gfloat dot = vec * norm_;
    if ( ret ) *ret = dot;
    return (dot > PROM_RTOL);
    }*/
  int numGhosts()const{
    assert(this!=NULL);
    return nodes_[0]->isGhost() + nodes_[1]->isGhost() + 
      nodes_[2]->isGhost() + nodes_[3]->isGhost();
  }
  void setMaster(int ism=TRUE){assert(this!=NULL);  bits_.pointer = ism?1:0;}
  bool isMaster() const { assert(this!=NULL); return bits_.pointer; }

  // data
  PromCoord    norm_;
private:
  PromGeoNode *nodes_private_[4];
  PromFacet   *neigh_private_[4];
  int          face_;
};

//! class PromHex: a hexaheron 
/*!
  class PromHex: a hexaheron -- 8 node brick
*/
class PromHex : public PromPoly
{
public:
  PromHex( PromNode **pn ) { 
    nodes_ = nodes_private_;
    neigh_ = (PromPoly**)neigh_private_;
    for( int i = 0 ; i < 8 ; i++ ) nodes_[i] = pn[i];
    for( int j = 0 ; j < 6 ; j++ ) neigh_[j] = NULL;
    setType(PROM_SOLID);
    bits_.dead = 0;
  }  
  virtual void reset(){ 
    assert(this!=NULL);    PromPoly::reset();
    nodes_[0] = NULL;  nodes_[1] = NULL; nodes_[2] = NULL;  nodes_[3] = NULL;
    nodes_[4] = NULL;  nodes_[5] = NULL; nodes_[6] = NULL;  nodes_[7] = NULL;
    
    neigh_[0] = NULL; neigh_[1] = NULL; neigh_[2] = NULL;
    neigh_[3] = NULL; neigh_[4] = NULL; neigh_[5] = NULL;
  }
  const int * getIDArr()const{ return octID_[0]; }
  virtual int numNodes() const{ assert(this!=NULL); return 8;}
  virtual int numNeigh() const{ assert(this!=NULL); return 6;}
private:
  PromNode *nodes_private_[8];
  PromHex *neigh_private_[6];
};

//! class PromShell: a quadrilateral shell 
/*!
  class PromShell: a quadrilateral shell  -- 3D
*/
class PromShell : public PromPoly
{
public:
  PromShell( PromNode **pn ) { // [4]
    nodes_ = nodes_private_;
    neigh_ = (PromPoly**)neigh_private_;
    for( int i = 0 ; i < 4 ; i++ ) { nodes_[i] = pn[i]; neigh_[i] = NULL;}
    setType(PROM_PLANE);
    bits_.dead = 0;
  }
  virtual void reset(){
    assert(this!=NULL);    PromPoly::reset();
    nodes_[0] = NULL; nodes_[1] = NULL; nodes_[2] = NULL;  nodes_[3] = NULL;
    neigh_[0] = NULL; neigh_[1] = NULL; neigh_[2] = NULL;  neigh_[3] = NULL;
  }
  const int * getIDArr()const{ return quadID_[0]; }
  virtual int numNodes() const { assert(this!=NULL); return 4; }
  virtual int numNeigh() const { assert(this!=NULL); return 4; }
private:
  PromNode *nodes_private_[4];
  PromShell *neigh_private_[4];
};

//! class PromNPoly: arbitray polygon
/*!
  class PromNPoly: arbitray polygon
*/
class PromNPoly : public PromPoly
{
public:
  PromNPoly( PromNode **pn, const int nn );
  virtual ~PromNPoly();
  virtual void reset(){ 
    assert(this!=NULL);    PromPoly::reset();
    for(unsigned i = 0 ; i < bits_.shp_sz ; i++ ) nodes_[i] = NULL;
  }
  const int * getIDArr()const{ return NULL; }  
  virtual int numNodes() const{ return bits_.shp_sz;}
  virtual int numNeigh()const;
  int         SetNumNodes( const int nn );
private:
  PromNode  **nodes_private_;  
};

#endif //_ELEMENT__H__

#include "prom_inln.icc"
