//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: prom_bank.C,v 1.6 2004/08/27 19:59:36 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           prom_back.C
//
#include "prom_grid.hh"

#define PROM_MAX_PSZS 3

#undef __FUNCT__  
#define __FUNCT__ "PromBank::PromBank"
/* PromBank::PromBank *************************************************
 *
 *  - destructure
 */
PromBank::PromBank(int bsz /*= 1024*/ ) :
  tetBlocks_(NULL), currTet_(NULL), freeTet_(NULL),
  facetBlocks_(NULL), currFacet_(NULL), freeFacet_(NULL), 
  aggNodeBlocks_(NULL), currAggNode_(NULL), freeAggNode_(NULL),    
  geoNodeBlocks_(NULL), currGeoNode_(NULL), freeGeoNode_(NULL),
  symbGeoDataBlocks_(NULL), symbDataBlocks_(NULL), symbNodeArrBlocks_(NULL),
  nPSizes_(0), blocksize_(bsz ), numTetFree_(0), numActiveTets_(0), 
  numFacetFree_(0), numNodesFree_(0), clearSymbCalled(FALSE)
{
  for(int i=0;i<PROM_MAX_PSZS;i++){ 
    Psizes_[i]=0; currP_[i]=NULL; PBlocks_[i]=NULL;numPFree_[i]=0;
  }
} 

#undef __FUNCT__  
#define __FUNCT__ "PromBank::~PromBank"
PromBank::~PromBank()
{ 
 int ierr = DeleteAllCaches();  assert(ierr==0);
}

#undef __FUNCT__  
#define __FUNCT__ "PromBank::DeleteAllCaches"
/* PromBank::DeleteAllCaches *************************************************
 *
 *  - like destructure
 */
int PromBank::DeleteAllCaches()
{ 
  int ierr;

  // first get rid of extraneous stuff
  ierr = DeleteSymb(); CHKERRQ(ierr);

  // goeNodes
  if( geoNodeBlocks_ != NULL ){
    while( !geoNodeBlocks_->isEmpty() ) {
      PromGeoNode * gn = geoNodeBlocks_->removeHead();
      delete [] gn; 
    }
    delete geoNodeBlocks_; geoNodeBlocks_ = NULL;
  }
  freeGeoNode_ = NULL; currGeoNode_ = NULL; 
  // aggNodes
  if( aggNodeBlocks_ != NULL ){
    while( !aggNodeBlocks_->isEmpty() ) {
      PromAggNode *an = aggNodeBlocks_->removeHead();
      delete [] an; 
    }
    delete aggNodeBlocks_; aggNodeBlocks_ = NULL;
  }
  freeAggNode_ = NULL; currAggNode_ = NULL;     
  numNodesFree_ = 0;

  // elements
  ierr = DeleteTets(); CHKERRQ(ierr);
  ierr = DeleteFacets(); CHKERRQ(ierr);
  ierr = DeleteP(); CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::DeleteSymb"
/* PromBank::DeleteSymb() *************************************
 *
 *  - delete all symbolic data
 */
int PromBank::DeleteSymb()
{
  // symbData
  if( clearSymbCalled ) return 0; 
  clearSymbCalled = TRUE;

  if( symbDataBlocks_ != NULL ){
    while( !symbDataBlocks_->isEmpty() ) {
      PromSymbData *cs = symbDataBlocks_->removeHead();
      delete [] cs;
    }
    delete symbDataBlocks_; symbDataBlocks_ = NULL;
    // symbgeoData
    if( symbGeoDataBlocks_ != NULL ){
      assert(geoNodeBlocks_ != NULL);
      while( !symbGeoDataBlocks_->isEmpty() ) {
	PromSymbGeoData *cs = symbGeoDataBlocks_->removeHead();
	delete [] cs; 
      }
      delete symbGeoDataBlocks_;
      symbGeoDataBlocks_ = NULL;
    }
    else assert(aggNodeBlocks_ != NULL);
    //
    while( !symbNodeArrBlocks_->isEmpty() ) {
      PromNode **cs = symbNodeArrBlocks_->removeHead(); PetscFree(cs);
    }
    delete symbNodeArrBlocks_; symbNodeArrBlocks_ = NULL;
  }
  // clear references
  if( geoNodeBlocks_ != NULL ){
    Position pos = geoNodeBlocks_->getHeadPosition(); assert(pos!=NULL);
    while( pos ) {
      PromGeoNode *gn = geoNodeBlocks_->getNext(pos);
      for( int ii = 0 ; ii < blocksize_ ; ii++, gn++ ) {
	gn->symbGeoData_ = NULL; gn->symbData_ = NULL;
      }
    }
  }
  // aggNodes
  if( aggNodeBlocks_ != NULL ){
    Position pos = aggNodeBlocks_->getHeadPosition(); assert(pos!=NULL);
    while( pos ) {
      PromAggNode *an = aggNodeBlocks_->getNext(pos);
      for( int ii = 0 ; ii < blocksize_ ; ii++, an++ ) {
	an->symbData_ = NULL;
      }
    }
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::DeleteTets"
/* PromBank::DeleteTets() *************************************
 *
 *  - delete all tets
 */
int PromBank::DeleteTets()
{
  if( tetBlocks_ != NULL ){
    while( !tetBlocks_->isEmpty() ) {
      PromDTet *tet = tetBlocks_->removeHead();  
      delete [] tet;   
    }
    delete tetBlocks_; tetBlocks_ = NULL;
  }
  freeTet_ = NULL; numTetFree_ = 0; numActiveTets_ = 0; currTet_ = NULL;

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::DeleteFacets"
/* PromBank::DeleteFacets() *************************************
 *
 *  - delete all facets
 */
int PromBank::DeleteFacets()
{
  if( facetBlocks_ != NULL ) {
    while ( !facetBlocks_->isEmpty() ) {
      PromFacet* f = facetBlocks_->removeHead();
      delete [] f;
    }
    delete facetBlocks_; facetBlocks_ = NULL; 
  }
  freeFacet_ = NULL; numFacetFree_ = 0; currFacet_ = NULL; 

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::GetFreeTet"
/* PromBank::GetFreeTet() *************************************
 * 
 *  - private helpers to allocate an element
 *
 *  RETURN:
 *    - need to allocate nodes
 */
int PromBank::GetFreeTet( PromDTet** newel )
{
  // see if a good one
  if ( freeTet_ ) {    // keep free list up
    *newel = freeTet_; // good one will break
    freeTet_ = dynamic_cast<PromDTet*>(freeTet_->next_);
    (*newel)->next_ = NULL;  
  }
  else { 
    if ( numTetFree_ == 0 ) {
      assert(currTet_==NULL);
      // alloc new block
      currTet_ = new PromDTet[blocksize_]; CHKNEW( currTet_, blocksize_ );
      for( int ii = 0 ; ii < blocksize_ ; ii++ ) currTet_[ii].kill(); 
      if( !tetBlocks_ ) {
	tetBlocks_ = new PromList<PromDTet*>(10); CHKNEW(tetBlocks_,1);
      }
      tetBlocks_->addTail( currTet_ );
      numTetFree_ = blocksize_;
    }
    *newel = currTet_++;
    numTetFree_--;
    if(numTetFree_ == 0)currTet_ = NULL;
  }
  assert((*newel)->next_ == NULL);
  
  // keep data helpers uptodate
  numActiveTets_++; 

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::GetFreeFacet"
/* PromBank::GetFreeFacet() *************************************
 * 
 *  - private helpers to allocate an facets
 *
 *  SIDEEFFECT:
 *    -
 */
int PromBank::GetFreeFacet( PromFacet** newel )
{
  *newel = NULL; // flag

  if( numFacetFree_ == 0 ) {
    if ( freeFacet_ ) {
      *newel = freeFacet_;      assert(freeFacet_->isDead() );
      freeFacet_ = dynamic_cast<PromFacet*>(freeFacet_->next_);
      (*newel)->next_ = NULL;
    }
    else {
      assert(currFacet_==NULL);
      currFacet_ = new PromFacet[blocksize_];   CHKNEW(currFacet_, blocksize_);
      for( int ii = 0 ; ii < blocksize_ ; ii++ ) currFacet_[ii].kill(); 
      if( !facetBlocks_ ) {
	facetBlocks_ = new PromList<PromFacet*>(10); CHKNEW(facetBlocks_,1);
      }
      facetBlocks_->addTail( currFacet_ );
      numFacetFree_ = blocksize_;
    }
  }
  
  // give next if necc.
  if ( *newel == NULL ) {
    *newel = currFacet_++; // this is the return value
    numFacetFree_--;
    if( numFacetFree_ == 0 ) currFacet_ = NULL;
  }
  assert((*newel)->next_ == NULL);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::GetFreeGeoNode"
/* PromBank::GetFreeGeoNode() *************************************
 * 
 *  - private helpers to allocate an nodes
 *
 */
int PromBank::GetFreeGeoNode( PromGeoNode** newel )
{
  *newel = NULL; assert(currAggNode_==NULL);   
  
  if ( numNodesFree_ == 0 ) {
    assert( currGeoNode_ == NULL );
    if( freeGeoNode_ ) {
      *newel = freeGeoNode_;      assert(freeGeoNode_->isDead() );
      freeGeoNode_ = dynamic_cast<PromGeoNode*>(freeGeoNode_->next_);
      (*newel)->next_ = NULL;
    }
    else{
      currGeoNode_ = new PromGeoNode[blocksize_];CHKNEW(currGeoNode_,1);
      if( geoNodeBlocks_ == NULL ) {
	geoNodeBlocks_ = new PromList<PromGeoNode*>(10);
	CHKNEW(geoNodeBlocks_,1);
	assert( !clearSymbCalled ); 
	symbDataBlocks_ = new PromList<PromSymbData*>(10);
	CHKNEW(symbDataBlocks_,1);
	symbGeoDataBlocks_ = new PromList<PromSymbGeoData*>(10);
	CHKNEW(symbGeoDataBlocks_,1);
	symbNodeArrBlocks_ = new PromList<PromNode**>(10);
	CHKNEW(symbNodeArrBlocks_,1);
      }
      geoNodeBlocks_->addTail( currGeoNode_ );  numNodesFree_ = blocksize_;
      if( !clearSymbCalled ) { 
	PromSymbData *symbData = new PromSymbData[blocksize_];
	CHKNEW(symbData,1);
	PromSymbGeoData *geoData = new PromSymbGeoData[blocksize_];
	CHKNEW(geoData,1);
	symbDataBlocks_->addTail( symbData );  
	symbGeoDataBlocks_->addTail( geoData );
	//
	int ierr; PromNode **arrp;
	ierr = PetscMalloc(blocksize_*prom_table_size*sizeof(PromNode*), &arrp);
	CHKERRQ(ierr);
	symbNodeArrBlocks_->addTail( arrp );
	for( int ii = 0, jj = 0 ; ii < blocksize_ ; ii++, jj += prom_table_size ) {
	  currGeoNode_[ii].symbGeoData_ = geoData++;
	  currGeoNode_[ii].symbData_ = symbData++;
	  ierr = currGeoNode_[ii].symbData_->adjac.SetTable(&arrp[jj],prom_table_size);
	  CHKERRQ(ierr);
	}
      }
      for( int ii = 0 ; ii < blocksize_ ; ii++ ) currGeoNode_[ii].kill(); 
    }
  }

  // give next if necc.
  if ( *newel == NULL ) {
    *newel = currGeoNode_++; // this is the return value
    numNodesFree_--; 
    if(numNodesFree_ == 0) currGeoNode_ = NULL;
  }
  assert((*newel)->next_ == NULL);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::GetFreeAggNode"
/* PromBank::GetFreeAggNode() *************************************
 * 
 *  - private helpers to allocate an nodes

 */
int PromBank::GetFreeAggNode( PromAggNode** newel )
{
  *newel = NULL; assert(currGeoNode_==NULL); 
  
  if( numNodesFree_ == 0 ) {
    assert( currAggNode_ == NULL );
    if( freeAggNode_ ) {
      *newel = freeAggNode_;      assert(freeAggNode_->isDead() );
      freeAggNode_ = dynamic_cast<PromAggNode*>(freeAggNode_->next_);
      (*newel)->next_ = NULL;
    }
    else{
      currAggNode_ = new PromAggNode[blocksize_];
      CHKNEW(currAggNode_, blocksize_);
      if( aggNodeBlocks_ == NULL ) {
	aggNodeBlocks_ = new PromList<PromAggNode*>(10);
	CHKNEW(aggNodeBlocks_,1);
	assert( !clearSymbCalled ); 
	symbDataBlocks_ = new PromList<PromSymbData*>(10);
	CHKNEW(symbDataBlocks_,1);
	symbNodeArrBlocks_ = new PromList<PromNode**>(10);
	CHKNEW(symbNodeArrBlocks_,1);
      }
      aggNodeBlocks_->addTail( currAggNode_ ); numNodesFree_ = blocksize_;
      if( !clearSymbCalled ) {
	PromSymbData *symbData = new PromSymbData[blocksize_];
	CHKNEW(symbData,1);
	symbDataBlocks_->addTail( symbData ); 
	int ierr; PromNode **arrp;
	ierr = PetscMalloc(blocksize_*prom_table_size*sizeof(PromNode*), &arrp);
	CHKERRQ(ierr);
	symbNodeArrBlocks_->addTail( arrp );
	for( int ii = 0, jj = 0 ; ii < blocksize_ ; ii++, jj += prom_table_size ) {
	  currAggNode_[ii].symbData_ = symbData++;
	  ierr = currAggNode_[ii].symbData_->adjac.SetTable(&arrp[jj],prom_table_size);
	  CHKERRQ(ierr);
	}
      }
      for( int ii = 0 ; ii < blocksize_ ; ii++ ) currAggNode_[ii].kill();
    }
  }

  // give next if necc.
  if ( *newel == NULL ) {
    *newel = currAggNode_++; // this is the return value
    numNodesFree_--;
    if(numNodesFree_ == 0) currAggNode_ = NULL;
  }
  assert((*newel)->next_ == NULL);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::newTet"
/* PromBank::newTet *************************************
 *
 *  - constructor for a PromBank manages new Tet
 */
PromDTet *PromBank::newTet( PromNode* a, PromNode* b, PromNode* c, PromNode* d)
{
  PromDTet *ret;
  GetFreeTet( &ret );    assert( ret->isDead() );
  ret->reset( a, b, c, d ); // call constructor
  assert( !ret->isDead() );
  
  ret->calcTestData();    // determinate
  
  return ret;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::newFacet"
/* PromBank::newFacet() *************************************
 *
g */
PromFacet *PromBank::newFacet( PromNode* a, PromNode* b,
			       PromNode* c, PromNode* d /* = NULL */)
{
  PromFacet *ret;

  GetFreeFacet( &ret );      assert( ret->isDead() );
  ret->reset( a, b, c, d ); // call constructor
  assert( !ret->isDead() );

  return ret;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::newAggNode"
/* PromBank::newAggNode() *************************************
 *
 */
PromAggNode *PromBank::newAggNode()
{
  PromAggNode *ret;
  GetFreeAggNode( &ret ); 

  ret->kill(FALSE);   assert( !ret->isDead() );
  //ret->symbData_->adjac.setReduced( FALSE );
  
  return ret;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::newGeoNode"
/* PromBank::newGeoNode() *************************************
 *
 */
PromGeoNode *PromBank::newGeoNode()
{
  PromGeoNode *ret;
  
  GetFreeGeoNode( &ret ); 
  ret->kill(FALSE);   assert( !ret->isDead() );
  //ret->symbData_->adjac.setReduced( FALSE );

  return ret;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::Kill"
/* PromBank::Kill() *************************************
 *
 *  - delete a PromBank manages element, elem is still usable
 *
 *  - add node on free node list
 *
 */
int  PromBank::Kill( PromNode *e )
{
  assert( e != NULL && !e->isDead() );

  e->reset(); assert( e->isDead() ); 
  e->next_ = NULL;
  e->Die( this ); // call right 'Deallocate';
  assert( e->isDead() );

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::Deallocate(PromGeoNode)"
/* PromBank::Deallocate() **********************************************
 *
 *  - set free internally for reuse
 */
int  PromBank::Deallocate( PromGeoNode * const e )
{
  assert( geoNodeBlocks_ != NULL);

  if( freeGeoNode_ == NULL ) {
    freeGeoNode_ = e; 
    e->next_ = NULL;
  }
  else {
    e->next_ = freeGeoNode_;
    freeGeoNode_ = e;
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::Deallocate(PromAggNode)"
/* PromBank::Deallocate() **********************************************
 *
 *  - set free internally for reuse
 */ 
int  PromBank::Deallocate( PromAggNode * const e )
{
  assert( aggNodeBlocks_ != NULL);

  e->next_ = freeAggNode_;
  freeAggNode_ = e;
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::Kill"
/* PromBank::Kill() *************************************
 *
 *  - delete a PromBank manages element, elem is still usable
 *
 *  - add facet on free facet list
 *
 */
int  PromBank::Kill( PromFacet *e )
{
  assert( e != NULL && !e->isDead() );
  e->kill(); assert(e->next_ == NULL);
  e->mark( FALSE ); // marked on degenerate facets
  
  // set free
  Deallocate(e);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::Deallocate(PromFacet)"
/* PromBank::Deallocate() *************************************
 *
 *  - set free internally for reuse
 */
int  PromBank::Deallocate( PromFacet * const e )
{
  assert( facetBlocks_ != NULL);

  e->next_ = freeFacet_;
  freeFacet_ = e;

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::Kill"
/* PromBank::Kill() *************************************
 *
 *  - delete a PromBank manages element, elem is still usable
 *
 */
int PromBank::Kill( PromDTet *e )
{
  assert( e != NULL && !e->isDead() );  
  e->next_ = NULL;
  e->kill();  

  // set free
  Deallocate(e);
  
  numActiveTets_--; // diagnostic?
 
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::Deallocate(PromDTet)"
/* PromBank::Deallocate() *************************************
 *
 *  - set free internally for reuse
 */
int  PromBank::Deallocate( PromDTet * const elem )
{
  assert( tetBlocks_ != NULL);  assert( elem->isDead() );  
  
  elem->next_ = freeTet_;
  freeTet_ = elem;

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromBank::getNumTets"
//
int PromBank::getNumTets()
{
  return numActiveTets_;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::newP"
/* PromBank::newP() *************************************
 *
 */
shpfloat *PromBank::newP( int sz )
{
  shpfloat *ret;  
  GetFreeP( &ret, sz ); 
  return ret;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::GetFreeP"
/* PromBank::GetFreeP() *************************************
 * 
 *  - private helpers to allocate an ps
 *
 *  SIDEEFFECT:
 *    -
 */
int PromBank::GetFreeP( shpfloat **newel, int sz )
{
  *newel = NULL; // flag
  int ii;
  for( ii=0 ; ii < nPSizes_ ; ii++ ){
    if( Psizes_[ii] == sz ) {
      // have it
      if( !freeP_[ii].isEmpty() ){
	*newel = freeP_[ii].removeHead(); 
      }
      else if( numPFree_[ii] != 0 ) {
	*newel = currP_[ii]; // this is the return value
	numPFree_[ii]--;
	if( numPFree_[ii] == 0 ) currP_[ii] = NULL;
	else currP_[ii] += sz; // inc for next 
      }
      else {
	assert(currP_[ii]==NULL); 	assert( PBlocks_[ii] != NULL );
	currP_[ii] = new shpfloat[sz*blocksize_];   CHKNEW(currP_[ii],blocksize_);
	PBlocks_[ii]->addTail( currP_[ii] );
	numPFree_[ii] = blocksize_;
	return GetFreeP( newel, sz ); // recursive call (setup now)
      }
      return 0; // got it
    }
    PBlocks_[nPSizes_] = new PromList<shpfloat*>(10); CHKNEW(PBlocks_[nPSizes_],1);
    Psizes_[nPSizes_] = sz;    assert(currP_[ii]==NULL);
    currP_[ii] = new shpfloat[sz*blocksize_];   CHKNEW(currP_[ii],blocksize_);
    PBlocks_[ii]->addTail( currP_[ii] );
    numPFree_[ii] = blocksize_;
    nPSizes_++; assert(nPSizes_ <= PROM_MAX_PSZS);
    return GetFreeP( newel, sz ); // recursive call (setup now)
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::Deallocate(shpfloat)"
/* PromBank::Deallocate() *************************************
 *
 *  - set free internally for reuse
 */
int  PromBank::Deallocate( shpfloat* const elem, int sz )
{
  for(int ii=0 ; ii<nPSizes_ ; ii++){
    if( Psizes_[ii] == sz ){
      freeP_[ii].addHead(elem);
      return 0;
    }
  }
  assert(0); // error!
  return -1;
}

#undef __FUNCT__  
#define __FUNCT__  "PromBank::DeleteP"
/* PromBank::DeleteP() *************************************
 *
 *  - delete all Ps
 */
int PromBank::DeleteP()
{
  for( int ii=0 ; ii<nPSizes_ ; ii++ ) {
    if( PBlocks_[ii] != NULL ) {
      while ( !PBlocks_[ii]->isEmpty() ) {
	shpfloat* f = PBlocks_[ii]->removeHead();
	delete [] f;
      }
      delete PBlocks_[ii]; PBlocks_[ii] = NULL; 
    }
    freeP_[ii].removeAll(); numPFree_[ii] = 0; currP_[ii] = NULL; 
  }

  return 0;
}

