//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: prom_context.C,v 1.6 2006/01/26 22:29:39 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           prom_context.C
// 
#include "prom_grid.hh"

char PromContext::stateNames_[PROM_NUMSTATE+1][12]=
{"SELECTED","DELETED","UNKNOWN","NOTUSED","NUMSTATE"};
char PromContext::typeNames_[PROM_NUMTOPO+1][12]=
{"CORNER","EDGE","SURFACE","INTERIOR","NUMTOPO"};
int PromContext::prom_logs_[LOG_TOT+1];
int PromContext::prom_stages_[11];


int prom_intcompar(const void *a, const void *b){
  int aa = *(int*)a, bb = *(int*)b;
  return (aa < bb) ? -1 : (aa == bb) ? 0 : 1;
}

int prom_intcompar2(const void *a, const void *b){
  int aa = *(int*)a, bb = *(int*)b;
  return (aa > bb) ? -1 : (aa == bb) ? 0 : 1;
}

int prom_floatcompar(const void *a, const void *b){
  float aa = *(float*)a, bb = *(float*)b;
  return (aa < bb) ? -1 : (aa == bb) ? 0 : 1;
}

#undef __FUNCT__  
#define __FUNCT__ "PromContext::PrintMatInfo"
/* PromContext::PrintMatInfo ***********************************************
 * 
 *  INPUT:
 *  SIDE EFFECTS:
 *
 *  RETURN:
 *    - PETSc error code 
 */   
int PromContext::PrintMatInfo( const PromMatrix *AA, const int level, 
			       const int factor, const int mype, const int np )
{ 
  double  nz_sum, nz_max, average; 
  int M, m;
#if defined(PROM_USE_PETSC)  
  MatInfo info;   int ierr;
  ierr = MatGetInfo( AA->mat_, MAT_GLOBAL_MAX, &info ); CHKERRQ(ierr);
  nz_max = info.nz_used;
  ierr = MatGetInfo( AA->mat_, MAT_GLOBAL_SUM, &info ); CHKERRQ(ierr);
  nz_sum = info.nz_used; 
  average = nz_sum / (double)(np/factor);
#elif defined(PROM_USE_PETRA)
  // TODO
  nz_max = nz_sum = average = 0.;
#else
#error NEEED one!
#endif
  M = AA->getM(); m = AA->getm();
  if( mype == 0 ) { 
    PetscPrintf(PETSC_COMM_SELF,"[%d]%s%2d:lneq=%7d,gneq=%10d",mype,
		(level==0)?"MG   ":"Level",level, m, M);
    const PromCRMatrix *KK = dynamic_cast<const PromCRMatrix*>(AA); 
    if(KK!=NULL && KK->C_)PetscPrintf(PETSC_COMM_SELF,", CR: lneq=%4d, gneq=%6d",
				      KK->C_->getm(),KK->C_->getM());
    PetscPrintf(PETSC_COMM_SELF," ave=%4.2e Mnz,max/ave=%4.2e,%4d nz/row,factor=%3d\n",
		average/1.e6,nz_max/average,
		(int)(nz_sum/(double)M+.5),factor);
  }
  
  return 0;
}  

#undef __FUNCT__  
#define __FUNCT__ "PromLocalID::Archive"
/* PromLocalID::Archive() ***********************************************
 *
 *   - Save/Write to file 
 *
 *  INPUT:
 *    -  file: file to read/write
 *    -  type: READ, WRITE, PRINT(pretty debug output)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code.
 */
int PromLocalID::Archive( FILE *file, const PromArchiveType type )
{
  if(type == PROM_PRINT) fprintf( file, "<pid=%d: gid=%d>\n",lIndex_,gIndex_);
  else if( type == PROM_WRITE ) fprintf( file, "%d %d ", lIndex_, gIndex_);
  else fscanf( file, "%d %d ", &lIndex_, &gIndex_ );
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromCoord::Write"
/* PromCoord::Write() ***********************************************
 *
 *   - PromCoord functions
 *
 *  INPUT:
 *    - 
 *  SIDE EFFECTS:
 *  RETURN:
 */
int PromCoord::Write( int **buff )const
{ 
  gfloat *fbuff = (gfloat*)*buff;
  for(int i=0;i<3;i++) *fbuff++ = data_[i]; 
  *buff = (int*)fbuff;

  return 0;
} 
#undef __FUNCT__  
#define __FUNCT__ "PromCoord::Read"
int PromCoord::Read( int **buff )
{ 
  gfloat *fbuff = (gfloat*)*buff;
  for(int i=0;i<3;i++) data_[i] = *fbuff++;
  *buff = (int*)fbuff;
  
  return 0;
} 

#undef __FUNCT__  
#define __FUNCT__ "PromCoord::Archive"
int PromCoord::Archive( FILE *file/*=stderr*/, 
			const PromArchiveType type/*=PRINT*/ )
{
  //int ierr;
  if( type == PROM_PRINT ) {
    fprintf(file,"PromCoord(%le %le %le)\n",data_[0],data_[1],data_[2]);
  }
  else if( type == PROM_WRITE ) {
    fprintf( file, "%le %le %le ", data_[0], data_[1], data_[2]);
  }
  else { 
    assert( type == PROM_READ );
    fscanf( file, "%le %le %le ", &data_[0], &data_[1], &data_[2] );
  }
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromContext::getNextFactor"
/* PromContext::getNextFactor() ***********************************************
 *
 *   - machine (and operator!!!) depenent factor
 *
 *  INPUT:
 *    - opts
 *    - np
 *    - lastf:
 *    - gnodes:
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - next factor
 */ 
int PromContext::getNextFactor( const PromOptions &opts, const int np,
				const int lastf, const int geq )
{ 
  int nf;
  GetNextFactor( opts, np, lastf, geq, &nf );
  return nf;
}
/* PromContext::getNextFactor() ***********************************************
 * 
 *   - machine (and operator!!!) depenent factor
 *
 *  INPUT: 
 *    - lastf:
 *    - gnodes:
 *    - nf: next factor (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code.
 */    
int PromContext::GetNextFactor( const PromOptions &opts, 
				const int numprocs, const int lastf,
				const int eq, int *nf )
{
  //*nf = 2*lastf; return 0; //debug
  int ierr,factor,tt; 
  bool flag; 
  double aa;
  // det default parameters
  float A = .1, B = 200; // rest (eg, alpha or no PARCH)
  float maxeq_p = 10000;   
  // override w/ user provide parameters
  ierr=opts.GetOptionInt("-prometheus_preduce_base",&tt,&flag);CHKERRQ(ierr);
  if( flag && tt > 0 )  B = (float)tt;
  ierr = opts.GetOptionDouble("-prometheus_preduce_rate",&aa,&flag);
  CHKERRQ(ierr);
  if( flag && aa >= 0 )  A = (float)aa;
  ierr=opts.GetOptionInt("-prometheus_preduce_maxn",&tt,&flag);CHKERRQ(ierr);
  if( flag && tt > 0 )  maxeq_p = (float)tt;
  
  /* np */
  float C = -(float)eq, t = sqrt(B*B - 4*A*C), tnp = (t-B)/(2*A);
  if( (float)eq/tnp > maxeq_p ) tnp = (float)eq/maxeq_p;
  int np = (int)(tnp + 0.5), opt = (np<2) ? eq : eq/np;
  if( np >= numprocs ) factor = 1; 
  else if( np <= 1 ) factor = numprocs;
  else { 
    const int factfact = (int)((float)numprocs/(float)np/(float)lastf + 0.5);
    if( factfact < 1 ) factor = lastf; // do nothing -- should be 1
    else if( factfact*lastf >= numprocs ) factor = numprocs;
    else{
      // search for best, legal factor
      int f2 = (int)((float)numprocs/(float)np/(float)lastf), f1 = f2+1;
      if( f2 < 1 ) f2 = 1;
      // get to a legal factor
      while( numprocs%(f1*lastf) != 0 ) f1++;
      while( numprocs%(f2*lastf) != 0 ) f2--;
      // if under a node get even
      //while( f1*lastf < nlp && nlp%(f1*lastf) != 0 ) f1++;
      //while( f2*lastf < nlp && nlp%(f2*lastf) != 0 ) f2--;
      // 
      assert(numprocs%(f1*lastf)==0 && numprocs%(f2*lastf)==0);
      float n1 = numprocs/(f1*lastf), n2 = numprocs/(f2*lastf);
      float e1 = (float)eq/n1, e2 = (float)eq/n2; 
      float ff1 = e1/(float)opt, ff2 = (float)opt/e2;
      // reward a node 
      //if( nlp == (f1*lastf) ) ff1 *= .7;
      //if( nlp == (f2*lastf) ) ff2 *= .7;
      if( numprocs == (f1*lastf) && f2 != 1 ) ff2 *= .85;
      // pick best
      factor = ( (ff1 < ff2) ? f1 : f2 ) * lastf;
    }
  }

  *nf = factor;

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromContext::ScanArrMPI"
/* PromContext::ScanArrMPI() ***********************************************
 *
 *   - collect [0, n1, n1+n2, ... , n1+n2+...nn]
 *
 *  INPUT:
 *    - myn: number here
 *    - array[Comm.size+1]: [0, n1, n1+n2, ... , n1+n2+...nn] CREATED (out)
 *    - Comm: MPI communicator
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code.
 */ 
int PromContext::ScanArrMPI( int myn, int *&array, MPI_Comm comm )
{
  // proc_gnode - MPI - (send)
  int ii, sz, ierr;              assert(array==NULL);
  if( comm == MPI_COMM_NULL ) return 1;
  PetscFunctionBegin;  
  MPI_Comm_size( comm, &sz );
  ierr = PetscMalloc((sz+1)*sizeof(int),&array);          CHKERRQ(ierr);
  MPI_Allgather( &myn, 1, MPI_INT, array+1, 1, MPI_INT, comm );
  for(ii=array[0] = 0 ; ii < sz ; ii++) array[ii+1] = array[ii+1] + array[ii];
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "PromContext::CheckClearMPI"
/* PromContext::CheckClearMPI() ***********************************************
 *
 *   - check for no stray MPI messages
 *
 *  INPUT:
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromContext::CheckClearMPI( MPI_Comm comm, const int mype, 
				const int verbose )
{ 
  int flag;
  MPI_Status status;
  int rbuf[PROM_BSZ]; 

  if(verbose>1)PetscPrintf(comm,"[%d]%s WAIT\n",mype,__FUNCT__); 
  //PetscSleep(1);
  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, comm, &flag, &status); 
  if( flag ) {
    int j, n; 
    PetscPrintf(PETSC_COMM_SELF,"[%d]ERROR:stray message\n", mype );
    MPI_Recv(&rbuf,PROM_BSZ,MPI_INT,status.MPI_SOURCE,status.MPI_TAG,comm,
	     &status);
    MPI_Get_count( &status, MPI_INT, &n );
    PetscPrintf(PETSC_COMM_SELF,"[%d] TAG=%d, SRC=%d, SZ=%d:{",
		mype, status.MPI_TAG, status.MPI_SOURCE, n );
    for( j = 0 ; j < n && j < 100 ; j++ ) {
      PetscPrintf(PETSC_COMM_SELF,"%d, ",rbuf[j]);
    }
    PetscPrintf(PETSC_COMM_SELF,"}\n");
    SETERRQ(1, "hard STRAY MESSAGES"); 
  }
  // die if anyone else does, keep from getting mixed up with this
  MPI_Barrier(comm);  // debug
  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, comm, &flag, &status); 
  if(flag)SETERRQ(1, "a STRAY MESSAGES"); 
  MPI_Barrier(comm);  

  return 0;
}

////////////////////////////////////////////////////////////////////////
//
//  File i/o helpers
//
#undef __FUNCT__  
#define __FUNCT__ "PromContext::printArr"
/* PromContext::printArr ******************************************************
 *
 *   - array write helpers 
 *
 *   INPUT:
 *     - file: file to use
 *     - arr: array to write
 *     - n: size of 'arr'
 *     - header: string header with a '%d' for reads
 *     - step: number values per line (0 for no limit)
 *
 *   RETURN:
 *     - PETSc error code
 */ 
int PromContext::printArr( FILE *file, const int *arr, const int n,
			   const char *header, const int step)
{
  fprintf(  file, header, n );        
  for ( int j = 0 ; j < n ; j++ ) {
    if ( (step > 0 && !(j%step)) || (!step && !j) ){
      fprintf(  file, "\n");        
    }
    fprintf( file, " %d", arr[j] );
  }

  fprintf( file, "\n");            
  
  return 0;
}
int PromContext::printArr( FILE *file, const int *arr, const int n, 
			   const int aux, const char *header, const int step)
{ 
  fprintf(  file, header, n, aux );  
  for ( int j = 0 ; j < n ; j++ ) {
    if ( (step > 0 && !(j%step)) || (!step && !j) ) {
      fprintf(  file, "\n");  
    }
    fprintf( file, " %d", arr[j]);
  }
  fprintf( file, "\n" );          
  
  return 0;
}
int PromContext::printArr( FILE *file, const bool *arr, const int n, 
			   const char *header, const int step)
{ 
  fprintf(  file, header, n );        
  for ( int j = 0 ; j < n ; j++ ) {
    if ( (step > 0 && !(j%step)) || (!step && !j) ){
      fprintf(  file, "\n");        
    }
    fprintf( file, " %c", arr[j] ? '1' : '0' );
  }
  
  fprintf( file, "\n");            
  
  return 0;
}
int PromContext::printArr( FILE *file, const double *arr, const int n,
			   const char *header, const int step)
{
  fprintf(  file, header, n );      
  for ( int j = 0 ; j < n ; j++ ) {
    if ( (step > 0 && !(j%step)) || (!step && !j) ) {
      fprintf(  file, "\n");        
    }
    fprintf( file," %9.4e", arr[j] );
  }
  
  fprintf( file, "\n");            
  return 0;
}
int PromContext::printArr( FILE *file, const double *arr, const int n,
			   const int aux, const char *header, const int step)
{
  fprintf(  file, header, n, aux );      
  for ( int j = 0 ; j < n ; j++ ) {
    if ( (step > 0 && !(j%step)) || (!step && !j) ) {
      fprintf(  file, "\n");        
    }
    fprintf( file," %9.4e", arr[j] );
  }
  
  fprintf( file, "\n");            
  return 0;
}
int PromContext::printArr( FILE *file, const float *arr, const int n,
			   const char *header, const int step)
{
  fprintf(  file, header, n );      
  for ( int j = 0 ; j < n ; j++ ) {
    if ( (step > 0 && !(j%step)) || (!step && !j) ) {
      fprintf( file, "\n");        
    }
    fprintf( file," %9.4e", arr[j] ); 
  }
  fprintf( file, "\n");            
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromContext::scanArr"
/* PromContext::scanArr *******************************************************
 *
 *   - array read helpers, allocate arrays if necc.
 *
 *   INPUT:
 *     - file: file to use(input)
 *     - arr: array to read(output)
 *     - n: size of 'arr'(output)
 *     - header: string header with a '%d' for reads(input)
 *     - step: number values per line (0 for no limit)(input)
 *
 *   RETURN:
 *     - PETSc error code
 */
int PromContext::scanArr( FILE *file, int **arrarg, int *n,
			  const char *header, const int step)
{
  assert( n!=NULL );  int num, *arr, ierr;
  
  if ( 1 > fscanf( file, header, &num ) ){
    *arrarg = NULL; *n = 0;
  }
  else {
    if( num > 0 ) {
      if(!(*arrarg)){
	ierr = PetscMalloc(num*sizeof(int), arrarg); CHKERRQ(ierr);
      }
      arr = *arrarg;
      for( int j = 0 ; j < num ; j++ ) {
	if( (step > 0 && !(j%step)) || (!step && !j) ) fscanf( file, "\n");
	if(1>fscanf(file,"%d",arr++))SETERRQ(1,"Read Faliure (1)");
      }
    }
    fscanf( file, "\n");   
    *n = num;
  }

  return 0;
}
int PromContext::scanArr( FILE *file, int **arrarg, int *n, int *aux,
			  const char *header, const int step)
{
  assert( n!=NULL );  int num, *arr, ierr;

  if ( 1 > fscanf( file, header, &num, aux ) ){
    *arrarg = NULL; *n = 0; *aux = 0;
  }
  else {
    if( num > 0 ) {
      if ( !*arrarg ) {
	ierr = PetscMalloc(num*sizeof(int),arrarg); CHKERRQ(ierr);
      }
      arr = *arrarg;
      for( int j = 0 ; j < num ; j++ ) {
	if( (step > 0 && !(j%step)) || (!step && !j) ) fscanf( file, "\n");
	if(1>fscanf(file," %d",arr++))SETERRQ(1,"Read Faliure (2)");
      }
    }
    fscanf( file, "\n");
    *n = num;
  }
  
  return 0;
}
  
int PromContext::scanArr( FILE *file, double **arrarg, int *n,
			  const char *header, const int step)
{ 
  assert( n!=NULL ); int num,ierr; 
  
  if ( fscanf( file, header, &num ) < 1 ){
    *arrarg = NULL; *n = 0;
  }
  else {
    *n = num;
    if( num > 0 ) {
      if( *arrarg == 0 ){
	ierr = PetscMalloc(num*sizeof(double),arrarg);CHKERRQ(ierr);
      } 
      double *arr = *arrarg;
      for( int j = 0 ; j < num ; j++ ) {
	if( (step > 0 && !(j%step)) || (!step && !j) ) fscanf( file, "\n");
	if( fscanf(file," %le",arr++) < 1){SETERRQ(1,"Read Faliure (3)");}
      }
    }
    fscanf( file, "\n");
  }
  
  return 0;
}
int PromContext::scanArr( FILE *file, float **arrarg, int *n,
			  const char *header, const int step)
{
  assert( n!=NULL ); int num,ierr; float *arr;

  if ( 1 > fscanf( file, header, &num) ) {
    *arrarg = NULL; *n = 0;
  }
  else {
    if( num > 0 ) {  
      if ( !*arrarg ) {
	ierr = PetscMalloc(num*sizeof(float),arrarg); CHKERRQ(ierr);
      }
      arr = *arrarg;
      for( int j = 0 ; j < num ; j++ ) {
	if((step > 0 && !(j%step)) || (!step && !j) ) fscanf( file, "\n");
	if(1>fscanf(file," %e",arr++))SETERRQ(1,"Read Faliure (4)");
      }
    }
    fscanf( file, "\n");
    *n = num;
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromContext::WriteMat"
/* PromContext::WriteMat ******************************************************
 *
 *   - writes a file: name/level/.m file of a matrix
 *
 *   INPUT:
 *     - mat: mat matrix to wite to a file (matlab readable format)
 *     - name: base name for base of file name
 *     - level: interger to append to name
 *
 *   SIDE EFFECTS:
 *     - writes a file  
 *
 *   RETURN:
 *     - PETSc error code.
 */
int PromContext::WriteMat( const PromMatrix *mat, char *name, int level )
{
  FILE *file;
  char path[128]; 
  int ierr,np=mat->map_.comm_.np(),mype=mat->map_.comm_.mype();

  if( level >= 0 ) {
    if( np == 1 ) sprintf( path, "%s%d", name, level);
    else sprintf ( path , "%s%d_%d", name, level, mype);    
    if ( (file = fopen(path,"w")) == NULL ){
      strcat( path, " - Failed to open matrix file\n");
      SETERRQ(1,path);
    }
  }
  else {
    if( np == 1 ) sprintf( path, "%s", name);
    else sprintf ( path , "%s%d", name, mype);    
    if ( (file = fopen(path,"w")) == NULL ){
      strcat( path, " - Failed to open matrix file\n");
      SETERRQ(1,path);
    }
  }
  ierr = WriteMat_private( mat, file, 0, 0, FALSE ); CHKERRQ(ierr);
  fclose( file );

  if( mat->isComplex() ) {
    const PromComplexMatrix *Ki = dynamic_cast<const PromComplexMatrix*>(mat); 
    assert(Ki!=0);
    strcat( path, "_i");
    if ( (file = fopen(path,"w")) == NULL ){
      strcat( path, " - Failed to open matrix file\n");
      SETERRQ(1,path);
    }
    ierr = WriteMat_private_i( Ki, file, 0, 0, FALSE ); CHKERRQ(ierr);
    fclose( file );
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromContext::WriteMat_private"
/* PromContext::WriteMat_private**********************************************
 *
 *   - writes a file: name/level/.m file of a matrix
 *
 *   INPUT:
 *     - mat: mat matrix to wite to a file (matlab readable format)
 *     - 
 *     - 
 *
 *   SIDE EFFECTS:
 *     - writes a file  
 *
 *   RETURN:
 *     - PETSc error code.
 */
int PromContext::WriteMat_private( const PromMatrix *mat, FILE *file, 
				   const int ioff, 
				   const int joff, const bool trans )
{  
  int ierr,gidi,gidj,kk,first_row,last_row,ncols, *colpp, ii, jj;
  int II, JJ;
  ierr = mat->GetOwnershipRange( &first_row, &last_row );   CHKERRQ(ierr);
  double *arrdat;  

  for( gidi = first_row ; gidi < last_row ; gidi++ ){ 
    ierr = mat->GetRow( gidi, &ncols, &colpp, &arrdat );    CHKERRQ(ierr);
    for ( kk = 0 ; kk < ncols ; kk++ ){
      if( arrdat[kk] == 0.0 ) continue;
      gidj = colpp[kk];
      ii = gidi + 1; jj = gidj + 1;
      if( trans ) { II = jj+ioff; JJ = ii+joff; }
      else { JJ = jj+joff; II = ii+ioff; }
      fprintf(file,"%d %d %.14e\n",II, JJ, arrdat[kk]);
    }
    ierr = mat->RestoreRow( gidi, &ncols, &colpp, &arrdat); CHKERRQ(ierr);
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromContext::WriteMat_private_i"
/* PromContext::WriteMat_private_i ********************************************
 *
 *   - writes a file: name/level/.m file of a matrix
 *
 *   INPUT:
 *     - mat: mat matrix to wite to a file (matlab readable format)
 *
 *   RETURN:
 *     - PETSc error code.
 */
int PromContext::WriteMat_private_i( const PromComplexMatrix *mat, FILE *file, 
				     const int ioff, 
				     const int joff, const bool trans )
{  
  int ierr,gidi,gidj,kk,first_row,last_row,ncols, *colpp, ii, jj;
  int II, JJ;
  ierr = mat->GetOwnershipRange( &first_row, &last_row );   CHKERRQ(ierr);
  double *arrdat;  

  for( gidi = first_row ; gidi < last_row ; gidi++ ){ 
    ierr = mat->GetRow_i( gidi, &ncols, &colpp, &arrdat );    CHKERRQ(ierr);
    for ( kk = 0 ; kk < ncols ; kk++ ){
      if( arrdat[kk] == 0.0 ) continue;
      gidj = colpp[kk];
      ii = gidi + 1; jj = gidj + 1;
      if( trans ) { II = jj+ioff; JJ = ii+joff; }
      else { JJ = jj+joff; II = ii+ioff; }
      fprintf(file,"%d %d %.14e\n",II, JJ, arrdat[kk]);
    }
    ierr = mat->RestoreRow( gidi, &ncols, &colpp, &arrdat); CHKERRQ(ierr);
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromContext::WriteMetis" 
/* PromContext::WriteMetis *******************************************
 *
 *   - write metis input file (not correct for parmetis)
 *
 *   INPUT:
 *      - name: file name
 *      - xadj: edge weights
 *      - ewgts: edge weights
 *      - vwgts: vertex weights
 *      - numnodes:
 *      - numadj:
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromContext::WriteMetis(char *name, const int *xadj, const int *adjncy, 
			    const int *vwgts, const int *ewgts, 
			    PromNode **nodes, const int nNodes, 
			    const int numadjac ) 
{
  int xx, ii, kk, numadj;
  FILE *file; 
  const char * const fmt = (!vwgts && !ewgts) ? "0 " : (!vwgts && ewgts) ? "1 " : 
    (vwgts && !ewgts) ? "10" : "11";
  PromNode *curr, *n1;
  
  if ( !(file = fopen(name,"w") ) ) {
    SETERRQ(1, "failed to open 'metis.grf' in WriteMetis()\n"); 
  }
  for( kk = ii = numadj = 0; kk < 2 ; kk++ ){
    if( kk == 1 ) fprintf(file," %d %d %s\n",nNodes,numadj,fmt);
    for ( xx = 0 ; xx < nNodes ; xx++ ){
      if( kk == 1 ) if( vwgts ) fprintf( file, "%d ", vwgts[xx]);
      curr = nodes[xx]; 
      TablePos tpos = curr->symbData_->adjac.getHeadPosition();assert(tpos!=0);
      while( tpos ){ 
	n1 = curr->symbData_->adjac.getNext(tpos);
	if( !n1->isGhost() ){
	  assert(n1->index_() != curr->index_());
	  if( kk == 0 ) {
	    if( n1->index_() > curr->index_() ) numadj++;
	  }
	  else {
	    if( ewgts ) fprintf( file, "%d %d ", adjncy[ii] + 1, ewgts[ii]);
	    else fprintf( file, "%d ", adjncy[ii] + 1 );
	    ii++;
	  }
	}
      }
      if( kk == 1 ) fprintf( file, "\n");
    }
  }
  assert( numadjac == ii );
  fclose(file);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromStack::XXXChild"
/* PromStack::removeChild() *************************************
 *
 *  - remove a node from simple linked list
 *
 */
bool PromStack::removeChild( PromStack *n )
{
  assert(this!=NULL);
  PromStack **n1 = &next_;
  while ( *n1 ) {
    if ( *n1 == n ) {
      *n1 = (*n1)->next_;      
      n->next_ = NULL;
      return TRUE;
    }
    n1 = &(*n1)->next_;
  }

  return FALSE;  
}

/* PromStack::findChild() *************************************
 *
 *  - remove from simple linked list
 *
 */
bool PromStack::findChild( const PromStack *n )const
{
  assert(this!=NULL && n);
  for ( const PromStack *n1 = next_ ; n1 ; n1 = n1->next_ ) 
    if ( n1 == n ) return TRUE;
  return FALSE;
}

/* PromStack::removeChildren() *************************************
 *
 *  - destructive remove and add children of this element to the list
 *
 *  INPUT
 *    - list: list to put children
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - num added
 */  
int PromStack::removeChildren( PromList<PromStack*> *list )
{
  assert(this!=NULL);
  PromStack *n2, *n1 = next_;
  next_ = NULL;
  int i = 0;
  while ( (n2 = n1) != NULL ) {
    list->addTail( n2 );
    n1 = n1->next_;
    n2->next_ = NULL;
    i++;
  }
  return i;
}

/* PromStack::numChildren() *************************************
 *
 *  - destructive remove and add children of this element to the list
 *
 *  INPUT
 *  SIDE EFFECTS:
 *  RETURN:
 *    - any found
 */  
int PromStack::numChildren()const
{  
  assert(this!=NULL);
  PromStack *n1 = next_;
  int i = 0;
  while ( n1 ) {
    n1 = n1->next_;
    i++;
  }
  return i;
}

/* PromStack::ClearChildren() *************************************
 *
 *  - clear child list
 *
 *  INPUT
 *  SIDE EFFECTS:
 *  RETURN:
 *    - number of children
 */  
int   PromStack::ClearChildren()
{
  assert(this!=NULL);
  PromStack  *n2, *n1 = next_;
  next_ = NULL;
  int i = 0;
  while ( (n2 = n1) != NULL ) {
    n1 = n1->next_;
    n2->next_ = NULL;
    i++;
  }
  return i;
}

/* PromStack::GetChildren() *************************************
 *
 *  - add children of this element to the list
 *
 *  INPUT
 *    - list: list to put children
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - any found
 * */  
int PromStack::GetChildren( PromList<PromStack*> *list )const
{  
  assert(this!=NULL);
  PromStack *n1 = next_;
  int i = 0;
  while ( n1 ) {
    list->addTail( n1 );
    n1 = n1->next_;
    i++;
  }
  return i;
} 

////////////////////////////////////////////////////////////////////////
// geometic  helpers
//
/* PromContext::cross() ******************************************
 *
 *    -  non-normalized Cross product, 
 *
 *         (n1->n2) X (n1->n3) / ||*||
 *
 *   INPUT:
 *      - n1,n2,n3: nodes to use in cross prod
 *      - crs: return value - cross product
 *
 *   SIDE EFFECTS:
 *      - 20 flops
 *
 *  RETURN:
 *    - square of area of (n1-n2-n3)
 */
gfloat PromContext::cross( const PromCoord &n1, const PromCoord &n2,
			   const PromCoord &n3, PromCoord *retcrs/*=NULL*/)
{
  double aa[3] = { n2[0]-n1[0], n2[1]-n1[1], n2[2]-n1[2] };
  double bb[3] = { n3[0]-n1[0], n3[1]-n1[1], n3[2]-n1[2] };
  double buff[3];
  buff[0] = aa[1] * bb[2] - aa[2] * bb[1];
  buff[1] = aa[2] * bb[0] - aa[0] * bb[2];
  buff[2] = aa[0] * bb[1] - aa[1] * bb[0];

  if (retcrs != NULL) { 
    (*retcrs)[0] = (gfloat)buff[0]; 
    (*retcrs)[1] = (gfloat)buff[1]; 
    (*retcrs)[2] = (gfloat)buff[2];}

  return (gfloat)(buff[0]*buff[0] + buff[1]*buff[1] + buff[2]*buff[2]);
}


/* PromContext::crossDot() ******************************************
 *
 *   - non-normalized Cross product,
 *         (n1->n2) X (n1->n3) / ||*||  *  (n1->n4)
 *
 *   INPUT:
 *     -n1,n2,n3,n4: nodes to use in cross prod
 *     -retcrs: return value-cross product
 *
 *   SIDE EFFECTS:
 *     -20 flops
 * 
 *   RETURN:
 *     -6 * volume of tet
 */  
gfloat PromContext::crossDot(const PromCoord &n1,
			     const PromCoord &n2,
			     const PromCoord &n3,
			     const PromCoord &n4,
			     PromCoord *retcrs /*= NULL*/)
{
  double aa[3] = { n2[0]-n1[0], n2[1]-n1[1], n2[2]-n1[2] };
  double bb[3] = { n3[0]-n1[0], n3[1]-n1[1], n3[2]-n1[2] };
  double cc[3] = { n4[0]-n1[0], n4[1]-n1[1], n4[2]-n1[2] };
  double buff[3];

  buff[0] = ( aa[1] * bb[2] ) - ( aa[2] * bb[1] );
  buff[1] = ( aa[2] * bb[0] ) - ( aa[0] * bb[2] );
  buff[2] = ( aa[0] * bb[1] ) - ( aa[1] * bb[0] );

  if (retcrs) { 
    (*retcrs)[0] = (gfloat)buff[0]; 
    (*retcrs)[1] = (gfloat)buff[1]; 
    (*retcrs)[2] = (gfloat)buff[2];}
   
  return (gfloat)(buff[0]*cc[0] + buff[1]*cc[1] + buff[2]*cc[2]);
}
